/*
  This file is part of the Changedfiles system. 
  $Id: chfctl.c,v 1.5 2003/02/03 04:43:00 rocky Exp $

  This product includes software developed by Konrad Rieck, modified
  by Rocky Bernstein.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at
  your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
******************************************************************************/
/*
  This is the Changefile control tool. It uses a unused syscall defined
  in changefiles.h in order to communicate with the module.
  Most features can be toggles by using this tool when the 
  module is loaded. Recompilation is not needed. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/syscall.h>

#include "headers.h"

chf_config_t c;

/*
 * Print the current status of the loaded module.
 */
void print_config()
{
    printf("\nCurrent configuration of the Changefiles v%s module:\n"
	   "\n  - Features\n", VERSION);
    printf("    Debugging:               %s\n", F(c.debug));
    printf("\n");
}

/*
 * Print an error message according to the submited err_msg value.
 */
void print_error(int err_msg)
{
    printf("Error %d \n", err_msg);

    switch (err_msg) {
    case NO_AUTH:
	printf("You are not authorized to use this program.\n");
	break;
    case NOT_LOADED:
	printf("Changefiles is not loaded. See modload(9E)\n");
	break;
    case INVALID_VALUE:
	printf("The assigned value is invalid for this type of variable\n");
	break;
    case INVALID_VARIABLE:
	printf("There is no such variable.\n");
	break;
    case MEM_FAULT:
	printf("Ex/Importing from user- to kernelspace failed.\n");
	break;
    case CONFIG_BLOCKED:
	printf("Configuration blocked. Volmgt running?\n");
	break;
    default:
	printf("Unknown error. *sigh*\n");
    }
    exit(EXIT_FAILURE);
}

void print_version(void)
{
    printf("Changefile v%s Control   "
	   "Copyright (c) 2002 Rocky Bernstein\n", VERSION);
}

/*
 * Guess what this does. 
 */
void print_usage()
{
    print_version();
    printf("Usage: chfctl -s variable=value [variable=value ...]\n"
	   "       chfctl -g | -v | -h\n"
	   "Options:\n"
	   "  -g                      get current configuration of the loaded module.\n"
	   "  -s variable=value ...   set current configuration of the loaded module.\n"
	   "  -q                      get free count in event queue.\n"
	   "  -O                      display overflow status.\n"
	   "  -h                      print this help.\n"
	   "  -v                      print version information.\n"
	   "\n"
	   " | Variable | Description                   | Values           |\n"
	   " |__________|_______________________________|__________________|\n"
	   " |    d     | debugging                     | off, on          |\n"
	   "\n"
	   "Example (Features and protections may not be compiled in):\n"
	   "  # chfctl -s d=off\n");
    exit(EXIT_FAILURE);
}

/*
 * Set a feature according to the given string. 
 */
int feature(char *variable)
{
    char *value;
    value = variable + 2;

    if (!strcmp(value, "on"))
	return ON;
    else if (!strcmp(value, "off"))
	return OFF;
    else {
	printf("Variable: %s\n", variable);
	print_error(INVALID_VALUE);
	return -1;
    }
}

/*
 * Parse the strings gathered from the commandline
 */
void parse_variable(char *variable)
{
    if (variable[1] != '=')
	print_usage();

    switch ((int) variable[0]) {
    case 'd':
	c.debug = feature(variable);
	break;
    default:
	printf("Variable: %s\n", variable);
	print_error(INVALID_VARIABLE);
    }
}


int main(int argc, char *argv[])
{
    int request = 0, i, count = 0, ret;
    char **variables = NULL;

    if (getuid())
	print_error(NO_AUTH);

    if (argc < 2)
	print_usage();
    else if (!strcmp(argv[1], "-g"))
	request = GET;
    else if (!strcmp(argv[1], "-s"))
	request = SET;
    else if (!strcmp(argv[1], "-q"))
	request = SIZE;
    else if (!strcmp(argv[1], "-O"))
	request = OVERFLOW;
    else if (!strcmp(argv[1], "-r")) {
      char action[1024]="";
      if (syscall(SYS_chfread, action))
	print_error(MEM_FAULT);
      printf("next action: %s\n", action);
      return (EXIT_SUCCESS);
    } else if (!strcmp(argv[1], "-v")) {
	print_version();
	return (EXIT_SUCCESS);
    } else
	print_usage();

    signal(SIGSYS, print_error);

    switch (request) {
    case GET:
      
      if (argc > 2)
	print_usage();
      if (syscall(SYS_chfcomm, &c, GET))
	print_error(MEM_FAULT);
      
      print_config();
      
      break;
    case SET:
      count = argc - 2;
      
      if (count < 1)
	print_usage();
      
      if (syscall(SYS_chfcomm, &c, GET))
	print_error(MEM_FAULT);
      
      variables = (char **) malloc(sizeof(char *) * count);
      for (i = 0; i < count; i++) {
	variables[i] = strdup(argv[i + 2]);
	parse_variable(variables[i]);
      }
      
      ret = syscall(SYS_chfcomm, &c, SET);
      
      if (ret)
	print_error(ret);
      
      print_config();
      break;
    case SIZE: 
      {
	char string[30];
	if (syscall(SYS_chfcomm, &string, SIZE))
	  print_error(MEM_FAULT);
	
	if (ret)
	  print_error(ret);
	
	printf("Queue size %s\n", string);
	break;
      }
    }

    return EXIT_SUCCESS;
}
