/*
 This file is part of the Changedfiles system. 
 $Id: comm.c,v 1.6 2003/02/08 04:16:57 rocky Exp $

 This product includes software developed by Konrad Rieck modified
 by Rocky Bernstein.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or (at
 your option) any later version.

 This program is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 USA
******************************************************************************/
/*
 * This file holds the Papillon communication syscall. The syscall is
 * protected by a read/write-lock. I know that's really strange for stuff
 * inside the kernel, but I guess the configuration is read a 1.000.000
 * times more often than written. 
 */

#define _KERNEL

#include "headers.h"
#include "libc.h"
#include "comm.h"
#include <sys/copyops.h>  /* copyinstr() */

/*
 * External configuration and rwlock. 
 */
extern chf_config_t config;

extern krwlock_t  config_rwlock;
extern kmutex_t   read_ready_lock;
extern kmutex_t   write_ready_lock;
extern kmutex_t   queue_lock;       /* Lock for event queue update */
extern kcondvar_t read_ready_cv;    /* Condition variable we signal on */
extern kcondvar_t write_ready_cv;   /* Condition variable we signal on */

#define BUF_SIZE        512
char *buffer[BUF_SIZE] = {NULL,};
int   buf_ov[BUF_SIZE] = {0,}; /* 1 if entry overwrote unread entry */
char *read_buf; 

int queue_head = 0;
int queue_tail = 0;
int queue_free = BUF_SIZE;

#define QUEUE_EMPTY queue_head == queue_tail

/*
 This is the communication control syscall, it only supports requests: GET
 and SET, SIZE, and OVERFLOW. 
*/
int64_t 
chf_comm(chf_config_t * c, enum request r)
{
    int64_t ret = 0;

    inc_refcnt();

    /*
     * If chf_comm is called from an unprivileged user or from inside a 
     * chroot environment, deny access.
     */
    if (CRED()->cr_ruid != SUSER_UID || u.u_rdir) {
	dec_refcnt();
	return set_errno(EPERM);
    }

    switch (r) {

    case SET:
	dcmn_err((CE_NOTE, "SET request for changedfiles configuration.\n"));

	/*
	 * This is the only case where we lock the config. Therefore
	 * we are usign a rwlock and not a mutex.
	 */
	if (!rw_tryenter(&config_rwlock, RW_WRITER)) {
	    ret = CONFIG_BLOCKED;
	    break;
	}

	if (copyin(c, &config, sizeof(config)))
	    ret = MEM_FAULT;

	rw_exit(&config_rwlock);

	break;

    case GET:
      dcmn_err((CE_NOTE, "GET request for changedfiles configuration.\n"));
      
      rw_enter(&config_rwlock, RW_READER);
      
      if (copyout(&config, c, sizeof(config)))
	ret = MEM_FAULT;
      
      rw_exit(&config_rwlock);
      break;
      
    case SIZE: {
      char int_str[30];
      unsigned int len;

      int_str[0]='\0';

      dcmn_err((CE_NOTE, "SIZE request. queue %d\n", queue_free));
      
      len = chf_uitostr(int_str, queue_free);
      rw_enter(&config_rwlock, RW_READER);
      
      if (copyout(&c, int_str, len+1))
	ret = MEM_FAULT;
      
      rw_exit(&config_rwlock);
      break;
    }
      
    default:
      ret = set_errno(EINVAL);
    }

    dec_refcnt();
    return ret;
}

/* Add one to the queue pointer wrapping around if needed */
static void inc_queue(int * queue_index) {
  (*queue_index)++;
  if (*queue_index ==  BUF_SIZE) *queue_index = 0;
}

/* 
   Return a string of the next file action that occured or "" if 
   nothing happened since last time. 
*/
int64_t 
chf_read(char *str)
{
    int64_t ret = 0;
    int len;

    /* Note to the writer that we've processed the last request. */ 
    mutex_enter(&write_ready_lock);
    cv_broadcast(&write_ready_cv);
    mutex_exit(&write_ready_lock);

    inc_refcnt();
    /*
     * If chf_comm is called from an unprivileged user or from inside a 
     * chroot environment, deny access.
     */
    if (CRED()->cr_ruid != SUSER_UID || u.u_rdir) {
	dec_refcnt();
	return set_errno(EPERM);
    }

    cmn_err(CE_CONT, "READ request %d\n", mutex_owned(&read_ready_lock));

    /* Block reader process until there's something to read...*/
    mutex_enter(&read_ready_lock);
#if 1
    /* Block for a limited time...*/
    while (QUEUE_EMPTY) {
      clock_t cur_ticks = ddi_get_lbolt();
      clock_t to = cur_ticks + drv_usectohz(60*1000000); /* 1 minute */
      if (cv_timedwait(&read_ready_cv, &read_ready_lock, to) == -1) {
	dcmn_err((CE_NOTE, "+++QUEUE READ TIMEOUT\n"));
	dec_refcnt();
	mutex_exit(&read_ready_lock);
	return 0;
      }
    }
#else 
    /* Block for an unlimited time...*/
    while (QUEUE_EMPTY) {
      cv_wait(&read_ready_cv, &read_ready_lock);
    }
#endif
    mutex_exit(&read_ready_lock);

    /* Pick out queue entry and leave. */
    mutex_enter(&queue_lock);
    if (buffer[queue_head] != NULL) {
      len = chf_strlen(buffer[queue_head])+1;
      dcmn_err((CE_NOTE, "len %d %s\n", len, buffer[queue_head])); 
      if (copyout(buffer[queue_head], str, len))
	ret = MEM_FAULT;
      
      kmem_free(buffer[queue_head], len);
    } else {
      dcmn_err((CE_NOTE, "buffer[%d] is NULL\n", queue_head));
    }
    inc_queue(&queue_head);
    queue_free++;
    mutex_exit(&queue_lock);

    dec_refcnt();

    return ret;
}

/* Add an action string to the end of the queue. 
   We prepend to this status giving whether or not this
   entry overwrote an existing entry (that wasn't read).
   0 = no overflow. 1 = overflow.
*/
#define STATUS_LEN 3
int64_t
chf_put_to_queue(const char * entry)
{
  int len;
  char ov_status[STATUS_LEN]; /* used to report back status */
  ov_status[0]='0';  /* Assume no overwrite. */
  ov_status[1]=' ';
  ov_status[2]='\0';

  dcmn_err((CE_NOTE, "put_to_queue: %s", entry)); 

  mutex_enter(&queue_lock);
  buf_ov[queue_tail] = (0 == queue_free);
  if ( buf_ov[queue_tail] ) {
    /* Full buffer. Overwrite oldest (head) entry. */
    inc_queue(&queue_head);
    queue_free++; /* decrement below will cancel this effect. */
    buf_ov[queue_tail]=1;
    ov_status[0]='1';
  } 
  len = chf_strlen(entry); 
  len++; 
  buffer[queue_tail] = kmem_zalloc(len+STATUS_LEN-1, KM_SLEEP); 
  if ( ! buffer[queue_tail] ) { 
    mutex_exit(&queue_lock);
    return -ENOMEM; 
  }
  chf_strncpy(buffer[queue_tail], ov_status, STATUS_LEN);
  chf_strncat(buffer[queue_tail], entry, len); 
  dcmn_err((CE_NOTE, "copied %d bytes: %s\n", len+STATUS_LEN-1, 
	    buffer[queue_tail])); 
  
  inc_queue(&queue_tail);
  queue_free--;
  mutex_exit(&queue_lock);
  dcmn_err((CE_NOTE, "---LEAVING WRITER QUEUE LOCK\n"));

  /* Hopefully there's a reader waiting for this event. */
  mutex_enter(&read_ready_lock);
  cv_broadcast(&read_ready_cv);
  mutex_exit(&read_ready_lock);
  dcmn_err((CE_NOTE, "---LEAVING PUT TO QUEUE\n"));

  /* Block process until reader has picked up request. 
     Future enhancement: block only if there's info to fill out.
   */
  mutex_enter(&write_ready_lock);
#if 1
  /* Block for an limited time...
     Future enhancement: block only if this request needs to be filled
     out.
   */
  while (!(QUEUE_EMPTY)) {
    clock_t cur_ticks = ddi_get_lbolt();
    clock_t to = cur_ticks + drv_usectohz(30*1000000); /* 30 seconds */
    if (cv_timedwait(&write_ready_cv, &write_ready_lock, to) == -1) {
      dcmn_err((CE_NOTE, "+++QUEUE WRITE TIMEOUT\n"));
      break;
    }
  }
#else 
  /* Block for an unlimited time...*/
  while (!(QUEUE_EMPTY)) {
    cv_wait(&write_ready_cv, &write_ready_lock);
  }
#endif
  mutex_exit(&write_ready_lock);

  return 1;
}

/* Free resources associated with the queue. */
void chf_queue_destroy(void) {
  mutex_enter(&queue_lock);
  for ( ; !(QUEUE_EMPTY); inc_queue(&queue_head)) {
    int len;
    if (buffer[queue_head]  != NULL) {
      len = chf_strlen(buffer[queue_head])+1;
      dcmn_err((CE_NOTE, "freeing %d bytes: %s\n", len, buffer[queue_head])); 
      kmem_free(buffer[queue_head], len);
    }
  }
  mutex_exit(&queue_lock);

}
